from collections.abc import Hashable
from dataclasses import dataclass, field
from types import EllipsisType
from typing import Annotated, Any, Literal

from typing_extensions import Doc

from pdag._utils import InitArgsRecorder


@dataclass(frozen=True)  # Frozen to be valid as a dictionary key
class ReferenceABC(InitArgsRecorder):
    """Represents a reference to a parameter or collection in a model."""

    name: Annotated[str, Doc("Name of the parameter or collection.")]

    # For time-series access
    previous: bool = field(default=False, kw_only=True)
    next: bool = field(default=False, kw_only=True)
    initial: bool = field(default=False, kw_only=True)
    all_time_steps: bool = field(default=False, kw_only=True)

    __init_args__: tuple[Any, ...] = field(init=False, compare=False, hash=False)
    __init_kwargs__: dict[str, Any] = field(init=False, compare=False, hash=False)

    def __post_init__(self) -> None:
        if sum([self.previous, self.next, self.initial, self.all_time_steps]) > 1:
            msg = "Reference cannot have more than one of previous, next, initial, or all_time_steps set."
            raise ValueError(msg)
        if not self.name.isidentifier():
            msg = "Reference name must be a valid identifier."
            raise ValueError(msg)

    @property
    def normal(self) -> bool:
        return not any([self.previous, self.next, self.initial, self.all_time_steps])


@dataclass(frozen=True)  # Frozen to be valid as a dictionary key
class ParameterRef(ReferenceABC):
    """Represents a reference to a parameter in a model."""


@dataclass(frozen=True)
class CollectionRef[K: Hashable](ReferenceABC):
    """Represents a reference to a collection in a model."""

    key: Annotated[K | None, Doc("Key to the collection. If None, it is a reference to the whole collection.")] = field(
        default=None,
    )


@dataclass(frozen=True)  # Frozen to be valid as a dictionary key
class MappingRef(CollectionRef[str | tuple[str | EllipsisType, ...]]):
    """Represents a reference to a mapping in a model."""


@dataclass(frozen=True)  # Frozen to be valid as a dictionary key
class ArrayRef(CollectionRef[tuple[int, ...]]):
    """Represents a reference to an array in a model."""


@dataclass(frozen=True)
class ExecInfo:
    """Represents the information available during the execution of a model."""

    attribute: Literal["n_time_steps", "time"]



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Hashable, Iterable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, ClassVar, Literal, cast

import numpy as np
import numpy.typing as npt

from pdag._utils import InitArgsRecorder

from .parameter import ParameterABC
from .reference import ArrayRef, CollectionRef, MappingRef
from .relationship import RelationshipABC

if TYPE_CHECKING:
    from types import EllipsisType


def key_to_str(key: Any, *, make_one_element_tuple_scalar: bool = False) -> str:
    if isinstance(key, tuple):
        if len(key) == 1 and make_one_element_tuple_scalar:
            return str(key[0])
        return ", ".join(str(k) for k in key)
    return str(key)


@dataclass
class CollectionABC[K: Hashable, T: ParameterABC[Any] | RelationshipABC](
    ABC,
    InitArgsRecorder,
):
    type: ClassVar[str] = "collection"
    name: str
    item_type: Literal["parameter", "relationship"] = field(init=False)

    def __post_init__(self) -> None:
        """Set the item type based on the first element in the collection and name the elements."""
        for element in self.values():
            if isinstance(element, ParameterABC):
                self.item_type = "parameter"
                break
            if isinstance(element, RelationshipABC):
                self.item_type = "relationship"
                break
            msg = "Collection must contain only parameters or relationships."
            raise TypeError(msg)
        self.name_elements()

    @abstractmethod
    def __getitem__(self, key: K) -> T:
        raise NotImplementedError

    @abstractmethod
    def items(self) -> Iterable[tuple[K, T]]:
        raise NotImplementedError

    def values(self) -> Iterable[T]:
        for _, element in self.items():  # noqa: PERF102
            yield element

    def keys(self) -> Iterable[K]:
        for key, _ in self.items():  # noqa: PERF102
            yield key

    def is_hydrated(self) -> bool:
        return all(elm.is_hydrated() for elm in self.values())

    def name_elements(self) -> None:
        for key, element in self.items():
            if element.name_is_set():
                # Already named
                continue
            element.name = f"{self.name}[{key_to_str(key)}]"

    def is_time_series(self) -> bool:
        if self.item_type == "parameter":
            return any(cast("ParameterABC[Any]", parameter).is_time_series for parameter in self.values())
        msg = "Only collections of parameters can be time series."
        raise TypeError(msg)

    @abstractmethod
    def ref(
        self,
        key: K | None = None,
        *,
        previous: bool = False,
        next: bool = False,  # noqa: A002
        initial: bool = False,
        all_time_steps: bool = False,
    ) -> CollectionRef[Any]:
        raise NotImplementedError


@dataclass
class Mapping[K: str | tuple[str, ...], T: ParameterABC[Any] | RelationshipABC](
    CollectionABC[K, T],
):
    """A collection of parameters or relationships that can be indexed by a string or tuple of strings."""

    type: ClassVar[str] = "mapping"
    name: str
    mapping: dict[K, T]

    def __getitem__(self, key: K) -> T:
        return self.mapping[key]

    def values(self) -> Iterable[T]:
        yield from self.mapping.values()

    def items(self) -> Iterable[tuple[K, T]]:
        yield from self.mapping.items()

    def keys(self) -> Iterable[K]:
        yield from self.mapping.keys()

    def ref(
        self,
        key: str | tuple[str | EllipsisType, ...] | None = None,
        *,
        previous: bool = False,
        next: bool = False,  # noqa: A002
        initial: bool = False,
        all_time_steps: bool = False,
    ) -> MappingRef:
        """Create a reference to the mapping."""
        return MappingRef(
            name=self.name,
            key=key,
            previous=previous,
            next=next,
            initial=initial,
            all_time_steps=all_time_steps,
        )


@dataclass
class Array[T: ParameterABC[Any] | RelationshipABC](CollectionABC[tuple[int, ...], T]):
    """A collection of parameters or relationships that can be indexed by a tuple of integers."""

    type: ClassVar[str] = "array"
    name: str
    array: npt.NDArray[T]  # type: ignore[type-var]

    @property
    def shape(self) -> tuple[int, ...]:
        return self.array.shape

    def __post_init__(self) -> None:
        super().__post_init__()

    def __getitem__(self, key: tuple[int, ...]) -> T:
        return self.array[key]  # type: ignore[no-any-return]

    def items(self) -> Iterable[tuple[tuple[int, ...], T]]:
        yield from np.ndenumerate(self.array)

    def ref(
        self,
        key: tuple[int, ...] | None = None,
        *,
        previous: bool = False,
        next: bool = False,  # noqa: A002
        initial: bool = False,
        all_time_steps: bool = False,
    ) -> ArrayRef:
        """Create a reference to the array."""
        return ArrayRef(
            name=self.name,
            key=key,
            previous=previous,
            next=next,
            initial=initial,
            all_time_steps=all_time_steps,
        )



from typing import ClassVar

from pdag._utils import MultiDef, multidef


class C(MultiDef):
    a: ClassVar[list[int]] = [1, 2]

    def f(self) -> list[int]:
        return self.a

    # In a loop, define multiple versions of 'g'.
    for i in range(3):

        @multidef(i)
        @staticmethod
        def g(i: int = i) -> int:
            return i


def test_multidef_access() -> None:
    c = C()

    assert c.get_def("g", 0)() == 0
    assert c.g[0]() == 0
    assert c.get_def("g", 0)(i=1) == 1
    assert c.g[0](i=1) == 1
    assert c.get_def("g", 1)() == 1
    assert c.g[1]() == 1
    assert c.get_def("g", 2)() == 2  # noqa: PLR2004
    assert c.g[2]() == 2  # noqa: PLR2004
    assert c.a == [1, 2]



"""Generate pages for the MkDocs documentation.

This script is called by the `gen-pages` MkDocs plugin to generate the index page and the API reference pages.
The index page is generated from the `README.md` file, and the API reference pages are generated from the source code.
"""

from pathlib import Path

import mkdocs_gen_files

PROJECT_ROOT = Path(__file__).parents[1]


def gen_index_page() -> None:
    readme_path = PROJECT_ROOT / "README.md"
    index_path = Path("index.md")

    readme_content = readme_path.read_text()
    index_content = adjust_documentation_paths(readme_content)
    index_content = remove_lines_with_documentation_link(index_content)
    index_content = convert_to_mkdocs_admonition(index_content)

    with mkdocs_gen_files.open(index_path, "w") as fd:
        fd.write(index_content)


def adjust_documentation_paths(content: str) -> str:
    """Adjust the documentation paths in the README.

    This function is used to adjust the documentation paths in the README for the index page.
    Example: `(docs/contents/cpu.md)` -> `(contents/cpu.md)`
    """
    return content.replace("(docs/", "(")


def remove_lines_with_documentation_link(content: str) -> str:
    """Remove lines with documentation links.

    This function is used to remove the lines with documentation links from the README for the index page.
    """
    return "\n".join(line for line in content.split("\n") if "(https://shunichironomura.github.io/pdag/)" not in line)


def convert_to_mkdocs_admonition(content: str) -> str:
    """Convert the GitHub-style alert to mkdocs admonition format.

    This function is used to convert the GitHub-style alert to mkdocs admonition format.
    Example:
    ```markdown
    > [!NOTE]
    > This is a note.
    > This is still a note.
    ```
    becomes:
    ```markdown
    !!! note
        This is a note.
        This is still a note.
    ```
    """
    lines = content.split("\n")

    alert_start_line_indices = [i for i, line in enumerate(lines) if line.startswith("> [!")]
    for start_index in alert_start_line_indices:
        end_index = start_index + 1
        while end_index < len(lines) and lines[end_index].startswith("> "):
            end_index += 1

        alert_lines = lines[start_index:end_index]
        alert_type = alert_lines[0][len("> [!") : alert_lines[0].index("]")]
        alert_content_lines = [line[2:] for line in alert_lines[1:]]
        lines[start_index:end_index] = [
            f"!!! {alert_type.lower()}",
            *(f"    {line}" for line in alert_content_lines),
        ]

    return "\n".join(lines)


def gen_ref_pages() -> None:
    nav = mkdocs_gen_files.Nav()  # type: ignore[attr-defined, no-untyped-call]

    src_dir = PROJECT_ROOT / "src"
    package_root = src_dir / "pdag"

    for path in sorted(package_root.rglob("*.py")):
        module_path = path.relative_to(src_dir).with_suffix("")
        doc_path = path.relative_to(src_dir).with_suffix(".md")
        full_doc_path = Path("reference", doc_path)

        parts = tuple(module_path.parts)

        if parts[-1] == "__init__":
            parts = parts[:-1]
            if any(part.startswith("_") for part in parts):
                # Skip private modules
                continue
            doc_path = doc_path.with_name("index.md")
            full_doc_path = full_doc_path.with_name("index.md")
        elif parts[-1] == "__main__" or any(part.startswith("_") for part in parts):
            # Skip main modules and private modules
            continue

        nav[parts] = doc_path.as_posix()

        with mkdocs_gen_files.open(full_doc_path, "w") as fd:
            identifier = ".".join(parts)
            print("::: " + identifier, file=fd)

        mkdocs_gen_files.set_edit_path(full_doc_path, path)

    with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
        nav_file.writelines(nav.build_literate_nav())


gen_index_page()

# Reference for the type parameter is not yet supported.
# See https://github.com/mkdocstrings/griffe/issues/342
gen_ref_pages()



"""A Python package that helps you create and execute a directed acyclic graph (DAG) of parameters and their relationships."""  # noqa: E501

__all__ = [
    "Array",
    "ArrayRef",
    "BooleanParameter",
    "CategoricalParameter",
    "CollectionABC",
    "CollectionRef",
    "CoreModel",
    "ExecInfo",
    "ExecutionModel",
    "FunctionRelationship",
    "Mapping",
    "MappingRef",
    "Model",
    "NodeId",
    "ParameterABC",
    "ParameterId",
    "ParameterRef",
    "RealParameter",
    "RelationshipABC",
    "RelationshipId",
    "StaticParameterId",
    "StaticRelationshipId",
    "SubModelRelationship",
    "TimeSeriesParameterId",
    "TimeSeriesRelationshipId",
    "create_exec_model_from_core_model",
    "distance_constrained_sampling",
    "execute_exec_model",
    "export_dot",
    "relationship",
    "results_to_df",
    "run_experiments",
    "sample_parameter_values",
    "sample_parameter_values",
]

from ._core import (
    Array,
    ArrayRef,
    BooleanParameter,
    CategoricalParameter,
    CollectionABC,
    CollectionRef,
    CoreModel,
    ExecInfo,
    FunctionRelationship,
    Mapping,
    MappingRef,
    ParameterABC,
    ParameterRef,
    RealParameter,
    RelationshipABC,
    SubModelRelationship,
)
from ._exec import (
    ExecutionModel,
    NodeId,
    ParameterId,
    RelationshipId,
    StaticParameterId,
    StaticRelationshipId,
    TimeSeriesParameterId,
    TimeSeriesRelationshipId,
    create_exec_model_from_core_model,
    execute_exec_model,
)
from ._experiment import distance_constrained_sampling, results_to_df, run_experiments, sample_parameter_values
from ._export import export_dot
from ._notation import (
    Model,
    relationship,
)



from typing import Annotated

import pdag


class SquareModel(pdag.Model):
    """Square model that squares a number."""

    # x is a real number parameter and is the input to the model
    x = pdag.RealParameter("x")

    # y is a real number parameter and is the output of the model
    y = pdag.RealParameter("y")

    # The relationship is defined as a static method
    # with the @pdag.relationship decorator
    @pdag.relationship
    @staticmethod
    def square(
        # The annotation `x.ref()` indicates that the value of `x` will be provided
        # as the value of the `x_arg` parameter when the model is executed.
        x_arg: Annotated[float, x.ref()],
        # The annotation `y.ref()` indicates that the return value of the method
        # will be assigned to the `y` parameter when the model is executed.
    ) -> Annotated[float, y.ref()]:
        """Square the input value."""
        return x_arg**2


if __name__ == "__main__":
    from rich import print  # noqa: A004

    core_model = SquareModel.to_core_model()
    exec_model = pdag.create_exec_model_from_core_model(core_model)
    results = pdag.execute_exec_model(
        exec_model,
        inputs={
            pdag.StaticParameterId((), "x"): 2.0,
        },
    )

    print(results)



from pdag._utils import merge_two_set_dicts


def test_merge_two_set_dicts() -> None:
    d1 = {"a": {1, 2}, "b": {3}}
    d2 = {"b": {4}, "c": {5, 6}}
    merged = merge_two_set_dicts(d1, d2)
    assert merged == {"a": {1, 2}, "b": {3, 4}, "c": {5, 6}}



import ast
import inspect
from collections.abc import Callable
from textwrap import dedent
from typing import Any

import asttokens


def _guess_indentation(source: str) -> str:
    """Guess the indentation of the source code."""
    lines = source.splitlines()
    if not lines:
        msg = "Source code is empty."
        raise ValueError(msg)
    # Find the first non-empty line that starts with a non-whitespace character.
    for line in lines:
        if (
            # Not empty
            line.strip()
            # Not a comment
            and not line.lstrip().startswith("#")
            # Indented
            and (line.startswith((" ", "\t")))
        ):
            break
    else:
        msg = "Cannot guess indentation: all lines are empty or comments or non-indented."
        raise ValueError(msg)
    # Guess the indentation based on the first non-empty line.
    return line[: len(line) - len(line.lstrip())]


def get_function_body(func: Callable[..., Any]) -> str:
    r'''Get the body of a function as a string.

    For example, given a function:
    ```python
    def sample_function() -> None:
        return None
    ```

    `get_function_body(sample_function)` will return:
    ```python
    "return None\n"
    ```

    !!! note
        Currently, it cannot extract the comment if it is the first line of the function body.

        For example, given a function:
        ```python
        def sample_function() -> None:
            # This is a comment
            return None
        ```

        `get_function_body(sample_function)` will return:
        ```python
        "return None\n"
        ```

        It can, however, extract the docstring. For example, given a function:
        ```python
        def sample_function() -> int:
            """This is a docstring."""
            return 1
        ```
        `get_function_body(sample_function)` will return:
        ```python
        '"""This is a docstring."""\nreturn 1\n'
        ```

    '''
    source = dedent(inspect.getsource(func))
    # Parse the source with asttokens to keep track of positions.
    atok = asttokens.ASTTokens(source, parse=True)

    # Locate the first FunctionDef node in the AST.
    function_node = None
    for node in ast.walk(atok.tree):  # type: ignore[arg-type]
        if isinstance(node, ast.FunctionDef):
            function_node = node
            break
    if function_node is None:
        msg = "Function definition not found in source."
        raise ValueError(msg)

    # Extract the source for each statement in the function body.
    start = atok.get_text_range(function_node.body[0], padded=False)[0]
    end = atok.get_text_range(function_node.body[-1], padded=False)[1]
    body = dedent(_guess_indentation(source) + source[start:end])
    if not body.endswith("\n"):
        body += "\n"
    return body



import warnings
from enum import Enum
from typing import Annotated, Any, Literal, cast

import numpy as np
import numpy.linalg as LA  # noqa: N812
import numpy.typing as npt
from typing_extensions import Doc

from pdag._core.parameter import BooleanParameter, CategoricalParameter, LiteralValueType, ParameterABC, RealParameter
from pdag._exec import ParameterId


def inverse_distance_real(
    parameter: RealParameter,
    distance: float,
    *,
    rng: np.random.Generator,
) -> float:
    """Inverse distance function for real parameters."""
    assert parameter.lower_bound is not None, f"Parameter {parameter} has no lower bound."
    assert parameter.upper_bound is not None, f"Parameter {parameter} has no upper bound."

    if "nominal" in parameter.metadata:
        nominal: float = parameter.metadata["nominal"]
    else:
        warnings.warn(f"Parameter {parameter} has no nominal value. Using the midpoint of the range.", stacklevel=2)
        nominal = (parameter.lower_bound + parameter.upper_bound) / 2

    # There are two possible values for one distance. We choose one randomly.
    if rng.uniform() < 0.5:  # noqa: PLR2004
        return nominal + distance * (parameter.upper_bound - nominal)
    return nominal - distance * (nominal - parameter.lower_bound)


def _calc_distance_real(
    parameter: RealParameter,
    value: float,
) -> float:
    assert parameter.lower_bound is not None, f"Parameter {parameter} has no lower bound."
    assert parameter.upper_bound is not None, f"Parameter {parameter} has no upper bound."

    if "nominal" in parameter.metadata:
        nominal: float = parameter.metadata["nominal"]
    else:
        warnings.warn(f"Parameter {parameter} has no nominal value. Using the midpoint of the range.", stacklevel=2)
        nominal = (parameter.lower_bound + parameter.upper_bound) / 2

    if value > nominal:
        return (value - nominal) / (parameter.upper_bound - nominal)
    return (nominal - value) / (nominal - parameter.lower_bound)


def _calc_distance_boolean(
    parameter: BooleanParameter,
    value: bool,  # noqa: FBT001
) -> float:
    if value in parameter.metadata["distance"]:
        disatnce = parameter.metadata["distance"][value]
        assert isinstance(disatnce, float), f"Distance for value {value} is not a float."
        return disatnce
    warnings.warn(f"No distance defined for value {value}. Using 1.", stacklevel=2)
    return 1


def _calc_distance_categorical[T: LiteralValueType](
    parameter: CategoricalParameter[T],
    value: T,
) -> float:
    if value in parameter.metadata["distance"]:
        distance = parameter.metadata["distance"][value]
        assert isinstance(distance, float), f"Distance for value {value!r} is not a float."
        return distance
    warnings.warn(f"No distance defined for value {value!r}. Using 1.", stacklevel=2)
    return 1


def calc_distance[T](
    parameter: Annotated[ParameterABC[T], Doc("A Pdag parameter.")],
    value: T,
) -> float:
    if isinstance(parameter, RealParameter):
        assert isinstance(value, float), f"Value {value} is not a float."
        return _calc_distance_real(parameter, value)
    if isinstance(parameter, BooleanParameter):
        assert isinstance(value, bool), f"Value {value} is not a bool."
        return _calc_distance_boolean(parameter, value)
    if isinstance(parameter, CategoricalParameter):
        assert isinstance(value, int | bytes | str | bool | Enum | None), f"Value {value} is not a literal value."
        return _calc_distance_categorical(parameter, value)

    msg = f"Unsupported parameter type: {type(parameter)}"
    raise ValueError(msg)


def normalize(
    x: npt.NDArray[np.number[Any]],
    ord: None | float | Literal["fro", "nuc"] = None,  # noqa: A002
    axis: None | int = None,
) -> npt.NDArray[np.floating[Any]]:
    norm: npt.NDArray[np.floating[Any]] = LA.norm(x, ord=ord, axis=axis, keepdims=True)  # type: ignore[assignment]
    return np.where(norm == 0.0, x, x.astype(np.float64) / norm)


def distance_constrained_sampling(
    parameters: dict[ParameterId, ParameterABC[Any]],
    n_samples: int,
    *,
    budget: float,
    rng: np.random.Generator | None = None,
    uniform_in_distance: bool = True,
) -> list[dict[ParameterId, Any]]:
    if not all(isinstance(parameter, RealParameter) for parameter in parameters.values()):
        msg = "Only real parameters are supported."
        raise NotImplementedError(msg)

    n_parameters = len(parameters)
    rng = np.random.default_rng() if rng is None else rng
    direction_samples = normalize(rng.normal(size=(n_samples, n_parameters)), axis=-1)
    radius_samples_normalized = (
        rng.uniform(size=n_samples) if uniform_in_distance else np.pow(rng.uniform(size=n_samples), 1 / n_parameters)
    )
    radius_samples = budget * radius_samples_normalized
    distance_samples = np.abs(radius_samples[:, np.newaxis] * direction_samples)

    # TODO: Support other parameter types apart from RealParameter
    return [
        {
            parameter_id: inverse_distance_real(cast("RealParameter", parameter), distance, rng=rng)
            for (parameter_id, parameter), distance in zip(parameters.items(), distance_sample, strict=True)
        }
        for distance_sample in distance_samples
    ]



from typing import Annotated

import numpy as np

import pdag

from ._square import SquareModel


class PolynomialModel(pdag.Model):
    """Polynomial model to demonstrate how to include a model in another model.

    This model calculates a polynomial `y = a[0] + a[1] * x + a[2] * x^2`
    where a[0], a[1], and a[2] are coefficients of the polynomial.
    """

    a = pdag.Array("a", np.array([pdag.RealParameter(...) for _ in range(3)]))
    x = pdag.RealParameter("x")
    x_squared = pdag.RealParameter("x_squared")
    y = pdag.RealParameter("y")

    calc_square_term = SquareModel.to_relationship(
        "calc_square_term",
        inputs={SquareModel.x.ref(): x.ref()},
        outputs={SquareModel.y.ref(): x_squared.ref()},
    )

    @pdag.relationship
    @staticmethod
    def polynomial(
        *,
        a: Annotated[list[float], a.ref()],
        x: Annotated[float, x.ref()],
        x_squared: Annotated[float, x_squared.ref()],
    ) -> Annotated[float, y.ref()]:
        return a[0] + a[1] * x + a[2] * x_squared


if __name__ == "__main__":
    from rich import print  # noqa: A004

    core_model = PolynomialModel.to_core_model()
    print(core_model)
    exec_model = pdag.create_exec_model_from_core_model(core_model)
    print(exec_model)



from typing import Annotated

import pdag

from ._square import SquareModel


class TwoSquares(pdag.Model):
    """TwoSquares model to demonstrate mapping containing submodels.

    This model calculates `z = x^2 + y^2`.
    """

    x = pdag.RealParameter("x")
    x_squared = pdag.RealParameter("x_squared")
    y = pdag.RealParameter("y")
    y_squared = pdag.RealParameter("y_squared")
    z = pdag.RealParameter("z")

    calc_square_term = pdag.Mapping(
        "calc_square_term",
        {
            input_param.name: SquareModel.to_relationship(
                ...,
                inputs={SquareModel.x.ref(): input_param.ref()},
                outputs={SquareModel.y.ref(): output_param.ref()},
            )
            for (input_param, output_param) in [(x, x_squared), (y, y_squared)]
        },
    )

    @pdag.relationship
    @staticmethod
    def squares(
        *,
        x_squared: Annotated[float, x_squared.ref()],
        y_squared: Annotated[float, y_squared.ref()],
    ) -> Annotated[float, z.ref()]:
        return x_squared + y_squared


if __name__ == "__main__":
    from rich import print  # noqa: A004

    core_model = TwoSquares.to_core_model()
    print(core_model)
    exec_model = pdag.create_exec_model_from_core_model(core_model)
    print(exec_model)



import inspect
from collections.abc import Callable, Hashable
from types import EllipsisType
from typing import Literal, get_args, get_origin, overload

from typing_extensions import _AnnotatedAlias

from pdag._core import (
    ExecInfo,
    FunctionRelationship,
    ReferenceABC,
)
from pdag._utils import MultiDefProtocol, get_function_body, multidef


def _get_outputs_from_signature(
    sig: inspect.Signature,
) -> tuple[list[ReferenceABC], bool]:
    def single_annotation_to_ref(
        annotations: _AnnotatedAlias,
    ) -> ReferenceABC:
        args = get_args(annotations)
        return next(iter(arg for arg in args if isinstance(arg, ReferenceABC)))

    if get_origin(sig.return_annotation) is not tuple:
        parameter_name = single_annotation_to_ref(sig.return_annotation)
        return [parameter_name], True
    args = get_args(sig.return_annotation)
    return [single_annotation_to_ref(arg) for arg in args], False


def _get_inputs_from_signature(
    sig: inspect.Signature,
) -> dict[str, ReferenceABC | ExecInfo]:
    """Get the input references from the signature of a function.

    The references are extracted from the annotations of the function's parameters.
    Note that it is possible that an argument does not have a reference annotation.
    For example, users may bind the loop variable in a for loop as the default value of a function argument.
    In this case, we return None, and the argument will be ignored.

    Example:
    ```python
    class MyModel(pdag.Model):
        for k in ["a", "b"]:
            @pdag.relationship(identifier=k)
            @staticmethod
            def f(
                x: Annotated[floag, pdag.ParameterRef("x")],
                k: str = k, # k is bound to the loop variable
            ) -> Annotated[float, pdag.ParameterRef("y")]:
                pass
    ```

    """

    def _get_ref_from_annotation(
        annotations: _AnnotatedAlias,
    ) -> ReferenceABC | ExecInfo | None:
        args = get_args(annotations)
        try:
            return next(iter(arg for arg in args if isinstance(arg, ReferenceABC | ExecInfo)))
        except StopIteration:
            # No annotation of type ReferenceABC or ExecInfo found.
            return None

    return {
        param.name: ref
        for param in sig.parameters.values()
        if (ref := _get_ref_from_annotation(param.annotation)) is not None
    }


@overload
def relationship[**P, T](
    func: Callable[P, T],
    *,
    identifier: None = None,
    at_each_time_step: Literal[False] = False,
) -> FunctionRelationship[P, T]: ...


@overload
def relationship[**P, T](
    func: None = None,
    *,
    identifier: Hashable = None,
    at_each_time_step: bool = False,
) -> Callable[[Callable[P, T]], FunctionRelationship[P, T]]: ...


def relationship[**P, T](
    func: Callable[P, T] | None = None,
    *,
    identifier: Hashable = None,
    at_each_time_step: bool = False,
) -> (
    FunctionRelationship[P, T]
    | Callable[[Callable[P, T]], FunctionRelationship[P, T]]
    | Callable[[Callable[P, T]], MultiDefProtocol[Hashable, FunctionRelationship[P, T]]]
):
    """Decorate a function to mark it as a relationship."""

    def decorator(
        func: Callable[P, T],
        *,
        _relationship_name: str | EllipsisType | None = None,
    ) -> FunctionRelationship[P, T]:
        # Get the function's signature
        sig = inspect.signature(func)
        inputs = _get_inputs_from_signature(sig)
        outputs, output_is_scalar = _get_outputs_from_signature(sig)
        function_body = get_function_body(func)
        return FunctionRelationship(
            _name=func.__name__ if _relationship_name is None else _relationship_name,
            inputs=inputs,
            outputs=outputs,
            function_body=function_body,
            output_is_scalar=output_is_scalar,
            _function=func,
            at_each_time_step=at_each_time_step,
        )

    if func is not None:
        # If the decorator is used without parentheses, we need to return the decorator
        return decorator(func)

    if identifier is None:
        return decorator

    # If the identifier is provided, we need to return decorator to return multidef

    def decorator_for_multidef(
        func: Callable[P, T],
    ) -> MultiDefProtocol[Hashable, FunctionRelationship[P, T]]:
        return multidef(identifier)(decorator(func, _relationship_name=...))

    return decorator_for_multidef



from typing import Annotated

import pdag


def test_capture_plain_parameter() -> None:
    class Model(pdag.Model):
        x = pdag.RealParameter("x")

    assert Model.parameters() == {"x": pdag.RealParameter("x")}


def test_time_step_parameter() -> None:
    class Model(pdag.Model):
        x = pdag.RealParameter("x", is_time_series=True)

    assert Model.parameters() == {"x": pdag.RealParameter("x", is_time_series=True)}


def test_capture_plain_relationship() -> None:
    class Model(pdag.Model):
        x = pdag.RealParameter("x")
        y = pdag.RealParameter("y")

        @pdag.relationship
        @staticmethod
        def f(x_arg: Annotated[float, pdag.ParameterRef("x")]) -> Annotated[float, pdag.ParameterRef("y")]:
            return x_arg

    assert Model.relationships() == {
        "f": pdag.FunctionRelationship(
            "f",
            inputs={"x_arg": pdag.ParameterRef("x")},
            outputs=[pdag.ParameterRef("y")],
            output_is_scalar=True,
            function_body="return x_arg\n",
            _function=Model.f._function,  # noqa: SLF001
            at_each_time_step=False,
        ),
    }



"""Poliynomials example to demonstrate how to include a model in another model."""

import pdag
from pdag.examples import PolynomialModel


def test_model_name() -> None:
    assert PolynomialModel.name == "PolynomialModel"


def test_polynomial_model() -> None:
    core_model = PolynomialModel.to_core_model()
    exec_model = pdag.create_exec_model_from_core_model(core_model)
    results = pdag.execute_exec_model(
        exec_model,
        inputs={
            pdag.StaticParameterId((), "a[0]"): 1.0,
            pdag.StaticParameterId((), "a[1]"): 2.0,
            pdag.StaticParameterId((), "a[2]"): 3.0,
            pdag.StaticParameterId((), "x"): 4.0,
        },
    )

    assert results == {
        pdag.StaticParameterId((), "a[0]"): 1.0,
        pdag.StaticParameterId((), "a[1]"): 2.0,
        pdag.StaticParameterId((), "a[2]"): 3.0,
        pdag.StaticParameterId((), "x"): 4.0,
        pdag.StaticParameterId((), "x_squared"): 4**2,
        pdag.StaticParameterId((), "y"): 1 + 2 * 4 + 3 * 4**2,
        pdag.StaticParameterId(("calc_square_term",), "x"): 4.0,
        pdag.StaticParameterId(("calc_square_term",), "y"): 4**2,
    }



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from types import EllipsisType
from typing import Annotated, Any, ClassVar

from typing_extensions import Doc

from pdag._utils import InitArgsRecorder

from .reference import ParameterRef


@dataclass
class ParameterABC[T](InitArgsRecorder, ABC):
    """Abstract base class for all parameters."""

    type: ClassVar[
        Annotated[
            str,
            Doc(
                "A string to represent the parameter type. "
                "It is not used now, but may be used in the future for model export.",
            ),
        ]
    ] = "base"
    _name: Annotated[str | EllipsisType, Doc("Name of the parameter.")]
    is_time_series: Annotated[bool, Doc("Whether it is a time-series parameter.")] = field(default=False, kw_only=True)
    metadata: Annotated[
        dict[str, Any],
        Doc("A dictionary where you can store any parameter metadata. It is not used by pdag."),
    ] = field(
        default_factory=dict,
        kw_only=True,
    )

    def is_hydrated(self) -> bool:
        """Check if all required attributes are set."""
        return self.name_is_set()

    def name_is_set(self) -> bool:
        """Check if the name is set."""
        return isinstance(self._name, str)

    @property
    def name(self) -> str:
        """Property to get the name of the parameter.

        If you try to access it before it is set, it will raise a ValueError.
        """
        if isinstance(self._name, EllipsisType):
            msg = "Parameter name has not been set."
            raise ValueError(msg)  # noqa: TRY004
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @abstractmethod
    def get_type_hint(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def from_unit_interval(self, value: float) -> T:
        """Map a value from a unit interval `[0, 1]` to the parameter value space.

        This method should be implemented so that the uniform distribution in the unit interval
        corresponds to the "uniform" distribution in the parameter value space.

        !!! note
            We use the term "uniform" in a loose sense.
            We don't treat the resulting distribution as a probability distribution and
            the reason that the distribution should be as "uniform" as possible
            is to sample the parameter value space efficiently.
        """
        raise NotImplementedError

    def ref(
        self,
        *,
        previous: bool = False,
        next: bool = False,  # noqa: A002
        initial: bool = False,
        all_time_steps: bool = False,
    ) -> ParameterRef:
        """Create a reference to this parameter in the model definition."""
        return ParameterRef(
            name=self.name,
            previous=previous,
            next=next,
            initial=initial,
            all_time_steps=all_time_steps,
        )


@dataclass
class RealParameter(ParameterABC[float]):
    type: ClassVar[str] = "real"
    unit: str | None = None
    lower_bound: float | None = None
    upper_bound: float | None = None

    def get_type_hint(self) -> str:
        return "float"

    def from_unit_interval(self, value: float) -> float:
        """Linearly map a value from a unit interval `[0, 1]` to the parameter value space.

        0 corresponds to the lower bound and 1 corresponds to the upper bound.
        """
        if self.lower_bound is None or self.upper_bound is None:
            msg = f"Lower and upper bounds must be set to convert from unit interval. Parameter: {self.name}"
            raise ValueError(msg)
        return self.lower_bound + value * (self.upper_bound - self.lower_bound)


@dataclass
class BooleanParameter(ParameterABC[bool]):
    type: ClassVar[str] = "boolean"

    def get_type_hint(self) -> str:
        return "bool"

    def from_unit_interval(self, value: float) -> bool:
        return value >= 0.5  # noqa: PLR2004


type LiteralValueType = int | bytes | str | bool | Enum | None


@dataclass
class CategoricalParameter[T: LiteralValueType](ParameterABC[T]):
    type: ClassVar[str] = "categorical"
    categories: tuple[T, ...]

    def get_type_hint(self) -> str:
        return f"Literal[{', '.join(self._to_type_hint(category) for category in self.categories)}]"

    @staticmethod
    def _to_type_hint(category: LiteralValueType) -> str:
        if isinstance(category, Enum):
            return f"{category.__class__.__name__}.{category.name}"
        return repr(category)

    def from_unit_interval(self, value: float) -> T:
        return self.categories[int(value * len(self.categories))]



from collections.abc import Mapping as MappingABC
from types import EllipsisType
from typing import Any, ClassVar, cast

from pdag._core import (
    CollectionABC,
    CoreModel,
    FunctionRelationship,
    Mapping,
    ParameterABC,
    ReferenceABC,
    RelationshipABC,
    SubModelRelationship,
)
from pdag._utils.multidef import MultiDef, MultiDefMeta, MultiDefStorage


def _function_relationship_multidef_storage_to_mapping(
    storage_name: str,
    storage: MultiDefStorage[str | tuple[str, ...], FunctionRelationship[Any, Any]],
) -> Mapping[str | tuple[str, ...], FunctionRelationship[Any, Any]]:
    return Mapping(
        name=storage_name,
        mapping=dict(storage),
    )


class ModelMeta(MultiDefMeta):
    def __new__(
        metacls,
        name: str,
        bases: tuple[type[Any], ...],
        namespace: dict[str, Any],
    ) -> type:
        cls = super().__new__(metacls, name, bases, namespace)

        # Tell type checker that cls is essentially of type Model
        cls = cast("type[Model]", cls)

        cls.name = name

        # Add top-level collections
        cls.__pdag_collections__ = {
            collection_name: collection
            for collection_name, collection in namespace.items()
            if isinstance(collection, CollectionABC)
        }

        # Add top-level parameters
        # This should be called before adding relationships because relationships depend on parameters being hydrated
        cls.__pdag_parameters__ = {
            parameter_name: parameter
            for parameter_name, parameter in namespace.items()
            if isinstance(parameter, ParameterABC)
        }

        # Add MultiDefStorage instances (function relationships defined in loops) to __pdag_collections__
        cls.__pdag_collections__.update(
            {
                storage_name: _function_relationship_multidef_storage_to_mapping(storage_name, storage)
                for storage_name, storage in namespace.items()
                if isinstance(storage, MultiDefStorage)
            },
        )

        # Add top-level relationships (function/submodel relationships at the top level)
        cls.__pdag_relationships__ = {}
        for relationship_name, relationship in namespace.items():
            if isinstance(relationship, RelationshipABC):
                cls.__pdag_relationships__[relationship_name] = relationship

        return cls


class Model(MultiDef, metaclass=ModelMeta):
    name: ClassVar[str]
    __pdag_parameters__: dict[str, ParameterABC[Any]]
    __pdag_relationships__: dict[str, RelationshipABC]
    __pdag_collections__: dict[str, CollectionABC[Any, Any]]

    @classmethod
    def parameters(cls) -> dict[str, ParameterABC[Any]]:
        return cls.__pdag_parameters__

    @classmethod
    def collections(cls) -> dict[str, CollectionABC[Any, Any]]:
        return cls.__pdag_collections__

    @classmethod
    def relationships(cls) -> dict[str, RelationshipABC]:
        return cls.__pdag_relationships__

    @classmethod
    def to_core_model(cls) -> CoreModel:
        return CoreModel(
            name=cls.__name__,
            parameters=cls.parameters(),
            collections=cls.collections(),
            relationships=cls.relationships(),
        )

    @classmethod
    def to_relationship(
        cls,
        /,
        name: str | EllipsisType,
        *,
        inputs: MappingABC[ReferenceABC, ReferenceABC],
        outputs: MappingABC[ReferenceABC, ReferenceABC],
        at_each_time_step: bool = False,
    ) -> SubModelRelationship:
        return SubModelRelationship(
            _name=name,
            submodel_name=cls.name,
            inputs=dict(inputs),
            outputs=dict(outputs),
            _submodel=cls.to_core_model(),
            at_each_time_step=at_each_time_step,
        )



from typing import Annotated

import pdag


class EachSquaredModel(pdag.Model):
    """EachSquaredModel model that uses a mapping of parameters."""

    m = pdag.Mapping("m", {k: pdag.RealParameter(...) for k in ("a", "b", "c")})
    m_squared = pdag.Mapping("m_squared", {k: pdag.RealParameter(...) for k in ("a", "b", "c")})

    for k in ("a", "b", "c"):
        # You need to provide the identifier to distinguish each relationship
        @pdag.relationship(identifier=k)
        @staticmethod
        def square(
            # The annotation `m.ref(k)` indicates that the value of `m[k]` will be provided
            # when the model is executed.
            m_arg: Annotated[float, m.ref(k)],
            # The annotation `m_squared.ref(k)` indicates that the return value of the method
            # will be assigned to `m_squared[k]` when the model is executed.
        ) -> Annotated[float, m_squared.ref(k)]:
            return m_arg**2



from typing import Any

import pytest

import pdag
from pdag.examples import DiamondMdpModel, PolynomialModel, SquareModel, TwoSquares

MODELS = [
    DiamondMdpModel,
    PolynomialModel,
    SquareModel,
    TwoSquares,
]

N_TIME_STEPS: dict[type[pdag.Model], int] = {
    DiamondMdpModel: 5,
    PolynomialModel: 1,
    SquareModel: 1,
    TwoSquares: 1,
}

INPUTS = {
    DiamondMdpModel: {
        pdag.StaticParameterId((), "policy"): "left",
        pdag.TimeSeriesParameterId((), "location", 0): "start",
    },
    PolynomialModel: {
        pdag.StaticParameterId((), "a[0]"): 1.0,
        pdag.StaticParameterId((), "a[1]"): 2.0,
        pdag.StaticParameterId((), "a[2]"): 3.0,
        pdag.StaticParameterId((), "x"): 2.0,
    },
    SquareModel: {
        pdag.StaticParameterId((), "x"): 2.0,
    },
    TwoSquares: {
        pdag.StaticParameterId((), "x"): 4.0,
        pdag.StaticParameterId((), "y"): 5.0,
    },
}

RESULTS: dict[type[pdag.Model], dict[pdag.ParameterId, Any]] = {
    DiamondMdpModel: {
        pdag.StaticParameterId(model_path=(), name="policy"): "left",
        pdag.TimeSeriesParameterId(model_path=(), name="location", time_step=0): "start",
        pdag.TimeSeriesParameterId(model_path=(), name="reward", time_step=0): 0.0,
        pdag.TimeSeriesParameterId(model_path=(), name="action", time_step=0): "go_left",
        pdag.TimeSeriesParameterId(model_path=(), name="location", time_step=1): "left",
        pdag.TimeSeriesParameterId(model_path=(), name="reward", time_step=1): 0.0,
        pdag.TimeSeriesParameterId(model_path=(), name="action", time_step=1): "move_forward",
        pdag.TimeSeriesParameterId(model_path=(), name="location", time_step=2): "end",
        pdag.TimeSeriesParameterId(model_path=(), name="reward", time_step=2): 1.0,
        pdag.TimeSeriesParameterId(model_path=(), name="action", time_step=2): "none",
        pdag.TimeSeriesParameterId(model_path=(), name="location", time_step=3): "end",
        pdag.TimeSeriesParameterId(model_path=(), name="action", time_step=3): "none",
        pdag.TimeSeriesParameterId(model_path=(), name="reward", time_step=3): 0.0,
        pdag.TimeSeriesParameterId(model_path=(), name="location", time_step=4): "end",
        pdag.TimeSeriesParameterId(model_path=(), name="action", time_step=4): "none",
        pdag.TimeSeriesParameterId(model_path=(), name="reward", time_step=4): 0.0,
        pdag.StaticParameterId(model_path=(), name="cumulative_reward"): 1.0,
    },
    PolynomialModel: {
        pdag.StaticParameterId(model_path=(), name="a[2]"): 3.0,
        pdag.StaticParameterId(model_path=(), name="a[1]"): 2.0,
        pdag.StaticParameterId(model_path=(), name="x"): 2.0,
        pdag.StaticParameterId(model_path=(), name="a[0]"): 1.0,
        pdag.StaticParameterId(model_path=("calc_square_term",), name="x"): 2.0,
        pdag.StaticParameterId(model_path=("calc_square_term",), name="y"): 4.0,
        pdag.StaticParameterId(model_path=(), name="x_squared"): 4.0,
        pdag.StaticParameterId(model_path=(), name="y"): 17.0,
    },
    SquareModel: {
        pdag.StaticParameterId(model_path=(), name="x"): 2.0,
        pdag.StaticParameterId(model_path=(), name="y"): 4.0,
    },
    TwoSquares: {
        pdag.StaticParameterId(model_path=(), name="x"): 4.0,
        pdag.StaticParameterId(model_path=(), name="y"): 5.0,
        pdag.StaticParameterId(model_path=("calc_square_term[x]",), name="x"): 4.0,
        pdag.StaticParameterId(model_path=("calc_square_term[y]",), name="x"): 5.0,
        pdag.StaticParameterId(model_path=("calc_square_term[x]",), name="y"): 16.0,
        pdag.StaticParameterId(model_path=("calc_square_term[y]",), name="y"): 25.0,
        pdag.StaticParameterId(model_path=(), name="x_squared"): 16.0,
        pdag.StaticParameterId(model_path=(), name="y_squared"): 25.0,
        pdag.StaticParameterId(model_path=(), name="z"): 41.0,
    },
}


@pytest.mark.parametrize("model", MODELS)
def test_model_to_core_model(model: type[pdag.Model]) -> None:
    """Test that the model can be converted to a core model."""
    _core_model = model.to_core_model()


@pytest.mark.parametrize(("model", "n_time_steps"), N_TIME_STEPS.items())
def test_model_to_exec_model(model: type[pdag.Model], n_time_steps: int) -> None:
    """Test that the model can be converted to an exec model."""
    core_model = model.to_core_model()
    _exec_model = pdag.create_exec_model_from_core_model(core_model, n_time_steps=n_time_steps)


@pytest.mark.parametrize(
    ("model", "n_time_steps", "inputs", "results_expected"),
    [(model, N_TIME_STEPS[model], INPUTS[model], RESULTS[model]) for model in MODELS],
)
def test_model_to_exec_model_and_execute(
    model: type[pdag.Model],
    n_time_steps: int,
    inputs: dict[pdag.ParameterId, Any],
    results_expected: dict[pdag.ParameterId, Any],
) -> None:
    """Test that the model can be converted to an exec model and executed."""
    core_model = model.to_core_model()
    exec_model = pdag.create_exec_model_from_core_model(core_model, n_time_steps=n_time_steps)
    results = pdag.execute_exec_model(
        exec_model,
        inputs=inputs,
    )
    assert results == results_expected, f"Expected {results_expected}, but got {results}"



#!/usr/bin/env -S uv run
# /// script
# requires-python = ">=3.8"
# dependencies = [
#     "tomli; python_version < '3.11'",
#     "typer",
# ]
# ///

import sys
from pathlib import Path
from typing import NoReturn

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

import typer


def main(
    pyproject_path: Path,
) -> NoReturn:
    with pyproject_path.open("rb") as f:
        pyproject = tomllib.load(f)

    version_str = pyproject["project"]["version"]
    print(version_str)
    raise typer.Exit(0)


if __name__ == "__main__":
    typer.run(main)



import pytest

import pdag

from .cases import CASES_DC_NOTATION_TO_CORE_MODEL


@pytest.mark.parametrize(("dc_notation", "core_model"), CASES_DC_NOTATION_TO_CORE_MODEL)
def test_dataclass_notation_to_core_model(
    dc_notation: type[pdag.Model],
    core_model: pdag.CoreModel,
) -> None:
    # Compare CoreModel
    exported_core_model = dc_notation.to_core_model()
    assert exported_core_model == core_model



from typing import Annotated, Any, Self

from typing_extensions import Doc


class InitArgsRecorder:
    """A mixin class that records the arguments used to initialize the instance.

    The arguments can be retrieved later using the `get_init_args` and `get_init_kwargs` methods.
    """

    # These are not class variables, but instance variables
    # They are declared here to inform type checkers.
    __init_args__: Annotated[tuple[Any, ...], Doc("Arguments used to initialize the instance.")]
    __init_kwargs__: Annotated[dict[str, Any], Doc("Keyword arguments used to initialize the instance.")]

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        """Record the arguments used to initialize the instance."""
        # Create the instance first
        instance = super().__new__(cls)
        # Save the arguments for later inspection
        # This is a bit of a hack to avoid assigning to attributes
        # even if the inherited dataclass has frozen=True
        object.__setattr__(instance, "__init_args__", args)
        object.__setattr__(instance, "__init_kwargs__", kwargs)
        return instance

    def get_init_args(self) -> tuple[Any, ...]:
        """Return the arguments used to initialize the instance."""
        return self.__init_args__

    def get_init_kwargs(self) -> dict[str, Any]:
        """Return the keyword arguments used to initialize the instance."""
        return self.__init_kwargs__



from typing import Annotated, Literal

import pytest

import pdag
from pdag._utils import get_function_body


def sample_function() -> None:
    return None


def test_get_function_body() -> None:
    body = get_function_body(sample_function)
    assert body == "return None\n"


class SampleClass(pdag.Model):
    @staticmethod
    def sqrt(
        *,
        x_arg: Annotated[float, pdag.ParameterRef("x")],
        z_arg: Annotated[Literal["pos", "neg"], pdag.ParameterRef("z")],
    ) -> Annotated[float, pdag.ParameterRef("y")]:
        if z_arg == "pos":
            return float(x_arg**0.5)
        return -float(x_arg**0.5)


def test_get_function_body_method() -> None:
    body = get_function_body(SampleClass.sqrt)
    assert body == 'if z_arg == "pos":\n    return float(x_arg**0.5)\nreturn -float(x_arg**0.5)\n'


def sample_function_starting_with_comment() -> None:
    # This is a comment
    return None


@pytest.mark.xfail(reason="Cannot extract the comment if it is the first line of the function body.")
def test_get_function_body_starting_with_comment() -> None:
    body = get_function_body(sample_function_starting_with_comment)
    assert body == "# This is a comment\nreturn None\n"


def sample_function_with_comment() -> int:
    x = 0

    # This is a comment
    return x + 1


def test_get_function_body_with_comment() -> None:
    body = get_function_body(sample_function_with_comment)
    assert body == "x = 0\n\n# This is a comment\nreturn x + 1\n"


def sample_function_with_docstring() -> int:
    """This is a docstring."""
    return 1


def test_get_function_body_with_docstring() -> None:
    body = get_function_body(sample_function_with_docstring)
    assert body == '"""This is a docstring."""\nreturn 1\n'



__all__ = [
    "ExecutionModel",
    "NodeId",
    "ParameterId",
    "RelationshipId",
    "StaticParameterId",
    "StaticRelationshipId",
    "TimeSeriesParameterId",
    "TimeSeriesRelationshipId",
    "create_exec_model_from_core_model",
    "execute_exec_model",
]
from .core import execute_exec_model
from .model import (
    ExecutionModel,
    NodeId,
    ParameterId,
    RelationshipId,
    StaticParameterId,
    StaticRelationshipId,
    TimeSeriesParameterId,
    TimeSeriesRelationshipId,
)
from .to_exec_model import create_exec_model_from_core_model



from textwrap import dedent
from typing import Annotated, Literal

import pdag


class SquareRootModel(pdag.Model):
    x = pdag.RealParameter("x")
    y = pdag.RealParameter("y")
    z = pdag.CategoricalParameter("z", categories=("pos", "neg"))

    @pdag.relationship
    @staticmethod
    def sqrt(
        *,
        x_arg: Annotated[float, pdag.ParameterRef("x")],
        z_arg: Annotated[Literal["pos", "neg"], pdag.ParameterRef("z")],
    ) -> Annotated[float, pdag.ParameterRef("y")]:
        if z_arg == "pos":
            return float(x_arg**0.5)
        return -float(x_arg**0.5)


square_root_core_model = pdag.CoreModel(
    name="SquareRootModel",
    parameters={
        "x": pdag.RealParameter("x"),
        "y": pdag.RealParameter("y"),
        "z": pdag.CategoricalParameter("z", categories=("pos", "neg")),
    },
    collections={},
    relationships={
        "sqrt": pdag.FunctionRelationship(
            "sqrt",
            inputs={"x_arg": pdag.ParameterRef("x"), "z_arg": pdag.ParameterRef("z")},
            outputs=[pdag.ParameterRef("y")],
            output_is_scalar=True,
            function_body=dedent(
                """\
                if z_arg == "pos":
                    return float(x_arg**0.5)
                return -float(x_arg**0.5)
                """,
            ),
            at_each_time_step=False,
        ),
    },
)

_TECHNOLOGIES = ["tech1", "tech2", "tech3"]


class TechDevCompletionModel(pdag.Model):
    dev_start_time = pdag.Mapping(
        "dev_start_time",
        {tech: pdag.RealParameter(...) for tech in _TECHNOLOGIES},
    )
    dev_time = pdag.Mapping(
        "dev_time",
        {tech: pdag.RealParameter(...) for tech in _TECHNOLOGIES},
    )
    dev_completion_time = pdag.Mapping(
        "dev_completion_time",
        {tech: pdag.RealParameter(...) for tech in _TECHNOLOGIES},
    )

    for tech in _TECHNOLOGIES:

        @pdag.relationship(identifier=tech)
        @staticmethod
        def dev_completion_time_model(
            dev_start_time: Annotated[float, pdag.MappingRef("dev_start_time", tech)],
            dev_time: Annotated[float, pdag.MappingRef("dev_time", tech)],
        ) -> Annotated[float, pdag.MappingRef("dev_completion_time", tech)]:
            return dev_start_time + dev_time


tech_dev_completion_core_model = pdag.CoreModel(
    name="TechDevCompletionModel",
    parameters={f"dev_start_time[{tech}]": pdag.RealParameter(f"dev_start_time[{tech}]") for tech in _TECHNOLOGIES}  # type: ignore[arg-type]
    | {f"dev_time[{tech}]": pdag.RealParameter(f"dev_time[{tech}]") for tech in _TECHNOLOGIES}
    | {f"dev_completion_time[{tech}]": pdag.RealParameter(f"dev_completion_time[{tech}]") for tech in _TECHNOLOGIES},
    collections={
        "dev_start_time": pdag.Mapping(  # type: ignore[dict-item]
            "dev_start_time",
            {tech: pdag.RealParameter(f"dev_start_time[{tech}]") for tech in _TECHNOLOGIES},
        ),
        "dev_time": pdag.Mapping(  # type: ignore[dict-item]
            "dev_time",
            {tech: pdag.RealParameter(f"dev_time[{tech}]") for tech in _TECHNOLOGIES},
        ),
        "dev_completion_time": pdag.Mapping(  # type: ignore[dict-item]
            "dev_completion_time",
            {tech: pdag.RealParameter(f"dev_completion_time[{tech}]") for tech in _TECHNOLOGIES},
        ),
    },
    relationships={
        "dev_completion_time_model": pdag.FunctionRelationship(
            "dev_completion_time_model",
            inputs={
                "dev_start_time": pdag.MappingRef("dev_start_time"),
                "dev_time": pdag.MappingRef("dev_time"),
            },
            outputs=[pdag.MappingRef("dev_completion_time")],
            output_is_scalar=True,
            function_body=dedent(
                """\
                return dev_start_time + dev_time
                """,
            ),
            at_each_time_step=False,
        ),
    },
)

CASES_DC_NOTATION_TO_CORE_MODEL = [
    (SquareRootModel, square_root_core_model),
]



"""Examples module that contains example models.

!!! warning
    The examples are not intended to be used in production code and are provided for demonstration purposes only.
"""

__all__ = [
    "DiamondMdpModel",
    "EachSquaredModel",
    "PolynomialModel",
    "SquareModel",
    "TwoSquares",
    "UmbrellaModel",
]
from ._diamond_mdp import DiamondMdpModel
from ._each_squared import EachSquaredModel
from ._polynomials import PolynomialModel
from ._square import SquareModel
from ._squares import TwoSquares
from ._umbrella import UmbrellaModel



import importlib.resources
from collections import defaultdict
from pathlib import Path
from typing import Any

from jinja2 import BaseLoader, Environment, Template

from pdag._core.model import CoreModel
from pdag._exec.to_exec_model import _iter_submodels_recursively

from .dot import to_dot_graph, tuple_to_html_id


def template_factory() -> Template:
    template_text = importlib.resources.files("pdag._export").joinpath("assets/graph.j2").read_text(encoding="utf-8")

    env = Environment(loader=BaseLoader(), autoescape=True)
    return env.from_string(template_text)


def _model_path_to_model_id(model_path: tuple[str, ...]) -> str:
    return tuple_to_html_id(model_path) if model_path else "__root__"


def _model_id_to_model_container_id(model_id: str) -> str:
    return f"model-container-{model_id}"


def to_html(core_model: CoreModel) -> str:
    graph_data: dict[str, Any] = {}
    click_events_dd: defaultdict[str, list[dict[str, Any]]] = defaultdict(list)

    for model_path, submodel in _iter_submodels_recursively(core_model, include_self=True):
        model_id = _model_path_to_model_id(model_path)
        graph = to_dot_graph(submodel, model_element_id=model_id)
        svg_data: bytes = graph.create_svg()  # pyright: ignore[reportAttributeAccessIssue]
        svg_string = svg_data.decode("utf-8")
        model_container_id = _model_id_to_model_container_id(model_id)

        if model_path:
            parent_model_path = model_path[:-1]
            parent_model_id = _model_path_to_model_id(parent_model_path)
            parent_node_id = tuple_to_html_id((parent_model_id, model_path[-1]))
            parent_model_container_id = _model_id_to_model_container_id(parent_model_id)
            assert parent_model_container_id in graph_data
            click_events_dd[parent_model_container_id].append(
                {
                    "source_id": parent_node_id,
                    "target_graph_id": model_container_id,
                },
            )
        else:
            parent_model_container_id = None

        assert model_container_id not in graph_data
        graph_data[model_container_id] = {
            "id": model_container_id,
            "svg": svg_string,
            "parent_id": parent_model_container_id,
            "path_label": " > ".join(("(root)", *model_path)),
        }

    click_events: dict[str, list[dict[str, Any]]] = dict(click_events_dd)

    template = template_factory()
    return template.render(
        title=core_model.name,
        graphs=[{"id": k, "click_events": click_events.get(k, []), **v} for k, v in graph_data.items()],
    )


def export_html(core_model: CoreModel, path: Path) -> None:
    rendered_html = to_html(core_model)
    with path.open("w", encoding="utf-8") as f:
        f.write(rendered_html)



from collections.abc import Mapping


def merge_two_set_dicts[K, V](d1: Mapping[K, set[V]], d2: Mapping[K, set[V]]) -> dict[K, set[V]]:
    """Merge two dictionaries where each value is a set.

    Values for duplicate keys are merged (unioned).

    Example:
    ```python
    d1 = {"a": {1, 2}, "b": {3}}
    d2 = {"b": {4}, "c": {5, 6}}
    merged = merge_two_set_dicts(d1, d2)
    # merged is {"a": {1, 2}, "b": {3, 4}, "c": {5, 6}}
    ```

    """
    result = dict(d1)
    for key, value in d2.items():
        result[key] = result.get(key, set()) | value
    return result
